@@ -16,6 +16,12 @@ logger = logging.getLogger(__name__)
 
 PRIORITY_RE = re.compile(r'^Priority ([\d]+)')
 
+BILL_TYPES = (
+        ('Billable', "Billable"),
+        ('DoNotBill', "Do Not Bill"),
+        ('NoCharge', "No Charge"),
+    )
+
 
 class InvalidStatusError(Exception):
     pass
@@ -43,6 +49,8 @@ class SyncJob(models.Model):
     start_time = models.DateTimeField(null=False)
     end_time = models.DateTimeField(blank=True, null=True)
     entity_name = models.CharField(max_length=100)
+    synchronizer_class = models.CharField(max_length=100, blank=True,
+                                          null=True)
     added = models.PositiveIntegerField(null=True)
     updated = models.PositiveIntegerField(null=True)
     deleted = models.PositiveIntegerField(null=True)
@@ -62,8 +70,26 @@ class AvailableConnectWiseBoardManager(models.Manager):
 
 
 class ConnectWiseBoard(TimeStampedModel):
+
     name = models.CharField(max_length=255)
     inactive = models.BooleanField(default=False)
+    work_role = models.ForeignKey(
+        'WorkRole',
+        blank=True,
+        null=True,
+        on_delete=models.SET_NULL
+    )
+    work_type = models.ForeignKey(
+        'WorkType',
+        blank=True,
+        null=True,
+        on_delete=models.SET_NULL
+    )
+    bill_time = models.CharField(
+        max_length=50,
+        choices=BILL_TYPES,
+        blank=True, null=True
+    )
 
     objects = models.Manager()
     available_objects = AvailableConnectWiseBoardManager()
@@ -210,6 +236,10 @@ class Member(TimeStampedModel):
         blank=True, null=True, max_length=20,
         choices=LICENSE_CLASSES, db_index=True
     )
+    work_type = models.ForeignKey(
+        'WorkType', null=True, on_delete=models.SET_NULL)
+    work_role = models.ForeignKey(
+        'WorkRole', null=True, on_delete=models.SET_NULL)
 
     objects = models.Manager()
     regular_objects = RegularMemberManager()
@@ -289,7 +319,7 @@ class Company(TimeStampedModel):
         ordering = ('identifier', )
 
     def __str__(self):
-        return self.get_identifier() or ''
+        return self.name or ''
 
     def get_identifier(self):
         return self.identifier
@@ -408,8 +438,9 @@ class Calendar(models.Model):
                 return start_day, days
             start_day = (start_day + 1) % 7
             days += 1
-            if days > 7:
-                # Calendar has no hours on any day
+            if days >= 7:
+                # Calendar has no hours on any day. This case can also occur
+                # if a calendar has seven consecutive holidays.
                 return None, None
 
     def is_holiday(self, date):
@@ -434,25 +465,30 @@ class Calendar(models.Model):
         start_time = datetime.timedelta(hours=start.hour,
                                         minutes=start.minute)
 
-        # get sla minutes for first day
-        end_of_day = self.get_day_hours(False, start.weekday())
+        # Get sla minutes for first day
+        start_day_end_time = self.get_day_hours(False, start.weekday())
 
-        if end_of_day and \
+        if start_day_end_time and \
                 not self.is_holiday(timezone.now().astimezone(tz=None)):
-            end_of_day = datetime.timedelta(hours=end_of_day.hour,
-                                            minutes=end_of_day.minute)
+            end_of_day = datetime.timedelta(hours=start_day_end_time.hour,
+                                            minutes=start_day_end_time.minute)
+        else:
+            end_of_day = None
 
         if start.date() == end.date():
+
             if end_of_day:
-                end_time = datetime.timedelta(
-                    hours=end.hour,
-                    minutes=end.minute) if \
-                    self.get_day_hours(False, start.weekday()) > \
-                    end.time() else end_of_day
-                minutes = (end_time - start_time).total_seconds() / 60
+
+                end_time = min(start_day_end_time, end.time())
+                end_time_delta = datetime.timedelta(
+                    hours=end_time.hour,
+                    minutes=end_time.minute
+                )
+
+                minutes = (end_time_delta - start_time).total_seconds() / 60
             # return sla time between start and end of day/end time, or zero
             # if start and end was outside of work hours
-            return minutes if minutes >= 0 else 0
+            return max(minutes, 0)
         else:
             if end_of_day and \
                     not self.is_holiday(timezone.now().astimezone(tz=None)):
@@ -482,7 +518,9 @@ class Calendar(models.Model):
                     continue
 
                 minutes = (end_of_day - start_of_day).total_seconds() / 60
-                sla_minutes = sla_minutes + minutes
+                # 24 hour calendar, minutes are full day
+                todays_minutes = minutes if minutes >= 0 else 1440
+                sla_minutes = sla_minutes + todays_minutes
                 day_of_week = (day_of_week + 1) % 7
                 current = current + datetime.timedelta(days=1)
 
@@ -513,12 +551,17 @@ class Calendar(models.Model):
             return sla_minutes
 
     def next_phase_expiry(self, sla_hours, ticket):
+
         start = ticket.entered_date_utc.astimezone(tz=None)
 
         # Start counting from the start of the next business day if the
         # ticket was created on a weekend
         day_of_week, days = self.get_first_day(start)
 
+        if day_of_week is None and days is None:
+            ticket.sla_expire_date = None
+            return
+
         sla_minutes, days, minutes, sla_start = self.get_sla_start(
                                                sla_hours,
                                                ticket.minutes_waiting,
@@ -549,7 +592,9 @@ class Calendar(models.Model):
                 continue
 
             minutes = (end_of_day - start_of_day).total_seconds() / 60
-            sla_minutes = sla_minutes - minutes
+            # 24 hour calendar, minutes are full day
+            todays_minutes = minutes if minutes >= 0 else 1440
+            sla_minutes = sla_minutes - todays_minutes
 
         # sla_minutes went below zero so we know that day is the expiry
         # Add the minutes back to sla_minutes and add sla minutes to the
@@ -658,8 +703,8 @@ class ScheduleStatus(models.Model):
 
 class ScheduleEntry(models.Model):
     name = models.CharField(max_length=250, blank=True, null=True)
-    expected_date_start = models.DateTimeField(blank=True, null=True)
-    expected_date_end = models.DateTimeField(blank=True, null=True)
+    date_start = models.DateTimeField(blank=True, null=True)
+    date_end = models.DateTimeField(blank=True, null=True)
     done_flag = models.BooleanField(default=False)
 
     ticket_object = models.ForeignKey(
@@ -727,12 +772,6 @@ class TimeEntry(models.Model):
         ('ChargeCode', "Charge Code"),
         ('Activity', "Activity")
     )
-    BILL_TYPES = (
-        ('Billable', "Billable"),
-        ('DoNotBill', "Do Not Bill"),
-        ('NoCharge', "No Charge"),
-        ('NoDefault', "No Default")
-    )
 
     class Meta:
         verbose_name_plural = 'Time entries'
@@ -756,12 +795,34 @@ class TimeEntry(models.Model):
     internal_analysis_flag = models.BooleanField(default=False)
     resolution_flag = models.BooleanField(default=False)
 
+    email_resource_flag = models.BooleanField(default=False)
+    email_contact_flag = models.BooleanField(default=False)
+    email_cc_flag = models.BooleanField(default=False)
+    email_cc = models.CharField(blank=True, null=True, max_length=1000)
+
     charge_to_id = models.ForeignKey(
         'Ticket', blank=True, null=True, on_delete=models.CASCADE)
     company = models.ForeignKey(
         'Company', blank=True, null=True, on_delete=models.CASCADE)
     member = models.ForeignKey(
         'Member', blank=True, null=True, on_delete=models.CASCADE)
+    work_type = models.ForeignKey(
+        'WorkType', blank=True, null=True, on_delete=models.SET_NULL)
+    agreement = models.ForeignKey(
+        'Agreement', blank=True, null=True, on_delete=models.SET_NULL)
+
+    def get_entered_time(self):
+        if self.time_end:
+            entered_time = self.time_end
+        elif self.actual_hours:
+            # timedelta does not like decimals
+            minutes = int(self.actual_hours * 60)
+            entered_time = self.time_start + timezone.timedelta(
+                minutes=minutes)
+        else:
+            entered_time = self.time_start
+
+        return entered_time
 
 
 class AvailableBoardTeamManager(models.Manager):
@@ -852,6 +913,50 @@ class ProjectStatus(TimeStampedModel):
         return self.name
 
 
+class ProjectType(TimeStampedModel):
+    name = models.CharField(max_length=30)
+    default_flag = models.BooleanField(default=False)
+    inactive_flag = models.BooleanField(default=False)
+
+    class Meta:
+        ordering = ('name', )
+        verbose_name_plural = 'Project types'
+
+    def __str__(self):
+        return self.name
+
+
+class ProjectPhase(TimeStampedModel):
+    description = models.CharField(max_length=100)
+    scheduled_start = models.DateField(blank=True, null=True)
+    scheduled_end = models.DateField(blank=True, null=True)
+    actual_start = models.DateField(blank=True, null=True)
+    actual_end = models.DateField(blank=True, null=True)
+
+    bill_time = models.CharField(
+        max_length=50, choices=BILL_TYPES, blank=True, null=True)
+    notes = models.TextField(null=True, blank=False)
+    scheduled_hours = models.DecimalField(
+        blank=True, null=True, decimal_places=2, max_digits=9)
+    actual_hours = models.DecimalField(
+        blank=True, null=True, decimal_places=2, max_digits=9)
+    budget_hours = models.DecimalField(
+        blank=True, null=True, decimal_places=2, max_digits=9)
+
+    project = models.ForeignKey(
+        'Project', blank=True, null=True, on_delete=models.SET_NULL
+    )
+    board = models.ForeignKey(
+        'ConnectwiseBoard', blank=True, null=True, on_delete=models.SET_NULL
+    )
+
+    class Meta:
+        verbose_name_plural = 'Project phases'
+
+    def __str__(self):
+        return self.description
+
+
 class AvailableProjectManager(models.Manager):
     """
     Return only projects whose status closed field is False.
@@ -870,7 +975,14 @@ class Project(TimeStampedModel):
         blank=True, null=True, decimal_places=2, max_digits=9)
     scheduled_hours = models.DecimalField(
         blank=True, null=True, decimal_places=2, max_digits=9)
+    actual_start = models.DateField(blank=True, null=True)
+    actual_end = models.DateField(blank=True, null=True)
+    estimated_start = models.DateField(blank=True, null=True)
+    estimated_end = models.DateField(blank=True, null=True)
 
+    company = models.ForeignKey(
+        'Company', on_delete=models.SET_NULL, blank=True, null=True
+    )
     status = models.ForeignKey(
         'ProjectStatus', blank=True, null=True, on_delete=models.SET_NULL)
     manager = models.ForeignKey(
@@ -880,6 +992,12 @@ class Project(TimeStampedModel):
         related_name='project_manager',
         on_delete=models.SET_NULL
     )
+    type = models.ForeignKey(
+        'ProjectType',
+        blank=True,
+        null=True,
+        on_delete=models.SET_NULL
+    )
 
     objects = models.Manager()
     available_objects = AvailableProjectManager()
@@ -903,11 +1021,11 @@ class OpportunityStage(TimeStampedModel):
 
 class AvailableOpportunityStatusManager(models.Manager):
     """
-    Return only Opportunity Statuses whose closed field is False.
+    Return only Opportunity Statuses whose inactive field is False.
     """
     def get_queryset(self):
         return super().get_queryset().filter(
-            closed_flag=False,
+            inactive_flag=False
         )
 
 
@@ -980,7 +1098,8 @@ class Opportunity(TimeStampedModel):
                                           on_delete=models.SET_NULL)
     priority = models.ForeignKey('OpportunityPriority',
                                  on_delete=models.SET_NULL, null=True)
-    stage = models.ForeignKey('OpportunityStage', on_delete=models.CASCADE)
+    stage = models.ForeignKey('OpportunityStage', blank=True, null=True,
+                              on_delete=models.SET_NULL)
     status = models.ForeignKey('OpportunityStatus', blank=True, null=True,
                                on_delete=models.SET_NULL)
     secondary_sales_rep = models.ForeignKey(
@@ -999,6 +1118,9 @@ class Opportunity(TimeStampedModel):
     def __str__(self):
         return self.name
 
+    def get_descriptor(self):
+        return self.name
+
     def get_connectwise_url(self):
         params = dict(
             recordType='OpportunityFv',
@@ -1033,10 +1155,23 @@ class Opportunity(TimeStampedModel):
 
 
 class Ticket(TimeStampedModel):
+    SCHEDULE_ENTRY_TYPE = "S"
+
+    PROJECT_TICKET = 'ProjectTicket'
+    PROJECT_ISSUE = 'ProjectIssue'
+    SERVICE_TICKET = 'ServiceTicket'
+
     RECORD_TYPES = (
-        ('ServiceTicket', "Service Ticket"),
-        ('ProjectTicket', "Project Ticket"),
-        ('ProjectIssue', "Project Issue"),
+        (SERVICE_TICKET, "Service Ticket"),
+        (PROJECT_TICKET, "Project Ticket"),
+        (PROJECT_ISSUE, "Project Issue"),
+    )
+
+    BILL_TIME_TYPES = (
+        ('Billable', "Billable"),
+        ('DoNotBill', "Do Not Bill"),
+        ('NoCharge', "No Charge"),
+        ('NoDefault', "No Default")
     )
 
     RESPOND = 'respond'
@@ -1062,7 +1197,6 @@ class Ticket(TimeStampedModel):
 
     actual_hours = models.DecimalField(
         blank=True, null=True, decimal_places=2, max_digits=9)
-    agreement_id = models.IntegerField(blank=True, null=True)
     approved = models.NullBooleanField(blank=True, null=True)
     budget_hours = models.DecimalField(
         blank=True, null=True, decimal_places=2, max_digits=9)
@@ -1098,6 +1232,14 @@ class Ticket(TimeStampedModel):
                                  max_length=250, choices=SLA_STAGE,
                                  db_index=True)
     minutes_waiting = models.PositiveIntegerField(default=0)
+    bill_time = models.CharField(blank=True, null=True,
+                                 max_length=20, choices=BILL_TIME_TYPES)
+    automatic_email_cc_flag = models.BooleanField(default=False)
+    automatic_email_cc = models.CharField(blank=True,
+                                          null=True,
+                                          max_length=1000)
+    automatic_email_contact_flag = models.BooleanField(default=False)
+    automatic_email_resource_flag = models.BooleanField(default=False)
 
     board = models.ForeignKey(
         'ConnectwiseBoard', blank=True, null=True, on_delete=models.CASCADE)
@@ -1117,6 +1259,9 @@ class Ticket(TimeStampedModel):
     project = models.ForeignKey(
         'Project', blank=True, null=True, related_name='project_tickets',
         on_delete=models.CASCADE)
+    phase = models.ForeignKey(
+        'ProjectPhase', blank=True, null=True, related_name='phase_tickets',
+        on_delete=models.SET_NULL)
     status = models.ForeignKey(
         'BoardStatus', blank=True, null=True, related_name='status_tickets',
         on_delete=models.SET_NULL)
@@ -1135,6 +1280,9 @@ class Ticket(TimeStampedModel):
     sub_type_item = models.ForeignKey(
         'Item', blank=True, null=True, related_name='item_tickets',
         on_delete=models.SET_NULL)
+    agreement = models.ForeignKey(
+        'Agreement', blank=True, null=True, related_name='agreement_tickets',
+        on_delete=models.SET_NULL)
 
     class Meta:
         verbose_name = 'Ticket'
@@ -1150,6 +1298,9 @@ class Ticket(TimeStampedModel):
     def __str__(self):
         return '{}-{}'.format(self.id, self.summary)
 
+    def get_descriptor(self):
+        return self.summary
+
     def time_remaining(self):
         time_remaining = self.budget_hours
         if self.budget_hours and self.actual_hours:
@@ -1208,9 +1359,15 @@ class Ticket(TimeStampedModel):
     def update_cw(self):
         """
         Send ticket status or priority and closed_flag updates to ConnectWise.
+        As of version 2019.3, project tickets and issues need to be updated
+        using the project API endpoint.
         """
-        service_client = api.ServiceAPIClient()
-        return service_client.update_ticket(
+        if self.record_type in [self.PROJECT_TICKET, self.PROJECT_ISSUE]:
+            api_client = api.ProjectAPIClient()
+        else:
+            api_client = api.ServiceAPIClient()
+
+        return api_client.update_ticket(
             self.id, self.closed_flag, self.priority, self.status
         )
 
@@ -1233,6 +1390,23 @@ class Ticket(TimeStampedModel):
         return self.save(*args, **kwargs)
 
     def calculate_sla_expiry(self):
+        # We can't guarantee that entered_date_utc won't be null.
+        # In the case that it is null, reset the date fields to prevent
+        # incorrect SLA calculations.
+        if not self.entered_date_utc:
+            self.sla_expire_date = None
+            self.do_not_escalate_date = None
+            self.sla_stage = None
+            self.date_resplan_utc = None
+            self.date_responded_utc = None
+            return
+
+        if not self.status:
+            logger.error(
+                'Ticket {}-{}, does not have a status set. '
+                'Skipping SLA calculation.'.format(self.id, self.summary)
+            )
+            return
 
         if not self.sla:
             return
@@ -1381,6 +1555,37 @@ class OpportunityNote(TimeStampedModel):
             str(self.id), self.opportunity.id)
 
 
+class ActivityStatus(TimeStampedModel):
+    name = models.CharField(max_length=50)
+    default_flag = models.BooleanField(default=False)
+    inactive_flag = models.BooleanField(default=False)
+    spawn_followup_flag = models.BooleanField(default=False)
+    closed_flag = models.BooleanField(default=False)
+
+    class Meta:
+        ordering = ('name', )
+        verbose_name_plural = 'activity statuses'
+
+    def __str__(self):
+        return self.name
+
+
+class ActivityType(TimeStampedModel):
+    name = models.CharField(max_length=50)
+    points = models.IntegerField()
+    default_flag = models.BooleanField(default=False)
+    inactive_flag = models.BooleanField(default=False)
+    email_flag = models.BooleanField(default=False)
+    memo_flag = models.BooleanField(default=False)
+    history_flag = models.BooleanField(default=False)
+
+    class Meta:
+        ordering = ('name', )
+
+    def __str__(self):
+        return self.name
+
+
 class Activity(TimeStampedModel):
     name = models.CharField(max_length=250)
     notes = models.TextField(blank=True, null=True, max_length=2000)
@@ -1392,16 +1597,52 @@ class Activity(TimeStampedModel):
         'Opportunity', blank=True, null=True, on_delete=models.CASCADE)
     ticket = models.ForeignKey(
         'Ticket', blank=True, null=True, on_delete=models.CASCADE)
+    status = models.ForeignKey(
+        'ActivityStatus', blank=True, null=True, on_delete=models.SET_NULL)
+    type = models.ForeignKey(
+        'ActivityType', blank=True, null=True, on_delete=models.SET_NULL)
+    company = models.ForeignKey(
+        'Company', blank=True, null=True, on_delete=models.CASCADE)
+    agreement = models.ForeignKey(
+        'Agreement', blank=True, null=True, on_delete=models.SET_NULL)
 
     class Meta:
         verbose_name_plural = 'activities'
-        # ordering = ('opportunity', 'ticket')
 
     def __str__(self):
-        return self.get_identifier() or ''
+        return self.name or ''
 
-    def get_identifier(self):
-        return self.name
+    def save(self, *args, **kwargs):
+        """
+        Save the object.
+
+        If update_cw as a kwarg is True, then update ConnectWise with changes.
+        """
+        update_cw = kwargs.pop('update_cw', False)
+        super().save(*args, **kwargs)
+        if update_cw:
+            self.update_cw()
+
+    def update_cw(self):
+        """
+        Send activity status and closed_flag updates to ConnectWise.
+        """
+        sales_client = api.SalesAPIClient()
+        return sales_client.update_activity_status(
+            self.id, self.status
+        )
+
+    def get_connectwise_url(self):
+        params = dict(
+            recordType='ActivityFv',
+            recid=self.id,
+            companyName=settings.CONNECTWISE_CREDENTIALS['company_id']
+        )
+        return '{}/{}?{}'.format(
+            settings.CONNECTWISE_SERVER_URL,
+            settings.CONNECTWISE_TICKET_PATH,
+            urllib.parse.urlencode(params)
+        )
 
 
 class SalesProbability(TimeStampedModel):
@@ -1585,3 +1826,41 @@ class Item(TimeStampedModel):
 
     def __str__(self):
         return self.name
+
+
+class WorkType(TimeStampedModel):
+    name = models.CharField(max_length=50)
+    inactive_flag = models.BooleanField(default=False)
+    overall_default_flag = models.BooleanField(default=False)
+    bill_time = models.CharField(
+        max_length=50, choices=BILL_TYPES, blank=True, null=True
+    )
+
+    def __str__(self):
+        return self.name
+
+
+class WorkRole(TimeStampedModel):
+    name = models.CharField(max_length=50)
+    inactive_flag = models.BooleanField(default=False)
+
+    def __str__(self):
+        return self.name
+
+
+class Agreement(TimeStampedModel):
+    name = models.CharField(max_length=100)
+    agreement_type = models.CharField(max_length=50, null=True)
+    cancelled_flag = models.BooleanField(default=False)
+    bill_time = models.CharField(
+        max_length=50, choices=BILL_TYPES, blank=True, null=True
+    )
+    work_type = models.ForeignKey(
+        'WorkType', null=True, on_delete=models.SET_NULL)
+    work_role = models.ForeignKey(
+        'WorkRole', null=True, on_delete=models.SET_NULL)
+    company = models.ForeignKey(
+        'Company', null=True, on_delete=models.SET_NULL)
+
+    def __str__(self):
+        return '{}/{}'.format(self.agreement_type, self.name)
