@@ -37,7 +37,13 @@ def init_members():
 
 def init_tickets():
     mocks.service_api_tickets_call()
-    synchronizer = sync.TicketSynchronizer()
+    synchronizer = sync.ServiceTicketSynchronizer()
+    return synchronizer.sync()
+
+
+def init_project_tickets():
+    mocks.project_api_tickets_call()
+    synchronizer = sync.ProjectTicketSynchronizer()
     return synchronizer.sync()
 
 
@@ -65,7 +71,7 @@ def init_slapriorities():
     mocks.service_api_get_sla_priorities_call(
         fixtures.API_SERVICE_SLA_PRIORITY_LIST
     )
-    synchronizer = sync.SLAPrioritySychronizer()
+    synchronizer = sync.SLAPrioritySynchronizer()
     return synchronizer.sync()
 
 
@@ -108,6 +114,20 @@ def init_project_statuses():
     return synchronizer.sync()
 
 
+def init_project_types():
+    mocks.projects_api_get_project_types_call(
+        fixtures.API_PROJECT_TYPES)
+    synchronizer = sync.ProjectTypeSynchronizer()
+    return synchronizer.sync()
+
+
+def init_project_phases():
+    mocks.projects_api_get_project_phases_call(
+        fixtures.API_PROJECT_PHASE_LIST)
+    synchronizer = sync.ProjectPhaseSynchronizer()
+    return synchronizer.sync()
+
+
 def init_projects():
     mocks.project_api_get_projects_call(fixtures.API_PROJECT_LIST)
     synchronizer = sync.ProjectSynchronizer()
@@ -176,7 +196,7 @@ def init_opportunities():
 def init_schedule_types():
     mocks.schedule_api_get_schedule_types_call(
         fixtures.API_SCHEDULE_TYPE_LIST)
-    synchronizer = sync.ScheduleTypeSychronizer()
+    synchronizer = sync.ScheduleTypeSynchronizer()
     return synchronizer.sync()
 
 
@@ -194,6 +214,20 @@ def init_schedule_entries():
     return synchronizer.sync()
 
 
+def init_activity_statuses():
+    mocks.sales_api_get_activities_statuses_call(
+        fixtures.API_SALES_ACTIVITY_STATUSES)
+    synchronizer = sync.ActivityStatusSynchronizer()
+    return synchronizer.sync()
+
+
+def init_activity_types():
+    mocks.sales_api_get_activities_types_call(
+        fixtures.API_SALES_ACTIVITY_TYPES)
+    synchronizer = sync.ActivityTypeSynchronizer()
+    return synchronizer.sync()
+
+
 def init_activities():
     mocks.sales_api_get_activities_call(
         fixtures.API_SALES_ACTIVITIES)
@@ -227,3 +261,24 @@ def init_items():
         fixtures.API_ITEM_LIST)
     synchronizer = sync.ItemSynchronizer()
     return synchronizer.sync()
+
+
+def init_work_types():
+    mocks.time_api_get_work_types_call(
+        fixtures.API_WORK_TYPE_LIST)
+    synchronizer = sync.WorkTypeSynchronizer()
+    return synchronizer.sync()
+
+
+def init_work_roles():
+    mocks.time_api_get_work_roles_call(
+        fixtures.API_WORK_ROLE_LIST)
+    synchronizer = sync.WorkRoleSynchronizer()
+    return synchronizer.sync()
+
+
+def init_agreements():
+    mocks.finance_api_get_agreements_call(
+        fixtures.API_AGREEMENT_LIST)
+    synchronizer = sync.AgreementSynchronizer()
+    return synchronizer.sync()
